@@ -42,6 +42,7 @@
 #include "lm-internals.h"
 #include "lm-parser.h"
 #include "lm-sha.h"
+#include "lm-queue.h"
 #include "lm-connection.h"
 
 #define IN_BUFFER_SIZE 1024
@@ -79,6 +80,9 @@ struct _LmConnection {
 
 	/* Communication */
 	GIOChannel *io_channel;
+	guint       io_watch_in;
+	guint       io_watch_err;
+	guint       io_watch_hup;
 
 	LmCallback *open_cb;
 	LmCallback *close_cb;
@@ -87,7 +91,7 @@ struct _LmConnection {
 
 	LmCallback *disconnect_cb;
 
-	GQueue     *incoming_messages;
+	LmQueue    *incoming_messages;
 	GSource    *incoming_source;
 
 	gint        ref_count;
@@ -167,10 +171,28 @@ static GSourceFuncs incoming_funcs = {
 static void
 connection_free (LmConnection *connection)
 {
+	int i;
+
 	g_free (connection->server);
 
-	if (connection->io_channel) {
-		g_io_channel_unref (connection->io_channel);
+	/* Unref handlers */
+	for (i = 0; i < LM_MESSAGE_TYPE_UNKNOWN; ++i) {
+		GSList *l;
+
+		for (l = connection->handlers[i]; l; l = l->next) {
+			HandlerData *hd = (HandlerData *) l->data;
+			
+			lm_message_handler_unref (hd->handler);
+			g_free (hd);
+		}
+
+		g_slist_free (connection->handlers[i]);
+	}
+
+	g_hash_table_destroy (connection->id_handlers);
+	
+	if (connection->is_open) {
+		connection_do_close (connection);
 	}
 
 	g_free (connection);
@@ -229,51 +251,84 @@ connection_new_message_cb (LmParser     *parser,
 		    _lm_message_type_to_string (lm_message_get_type (m)),
 		    lm_message_node_get_attribute (m->node, "from"));
 
-	g_queue_push_tail (connection->incoming_messages, m);
+	lm_queue_push_tail (connection->incoming_messages, m);
 }
 
 static gboolean
 connection_do_open (LmConnection *connection, GError **error)
 {
-	struct hostent     *he;
-        struct in_addr     *haddr;
-        struct sockaddr_in  saddr;
-	gint                fd;
+	gint             fd = -1;
+	int              err = -1;
+	struct addrinfo  req;
+	struct addrinfo *ans;
+	struct addrinfo *tmpaddr;
+	char             name[NI_MAXHOST];
+	char             portname[NI_MAXSERV];
 	
 	g_return_val_if_fail (connection != NULL, FALSE);
 	
-        he = gethostbyname(connection->server);
-        if (he == NULL) {
- 		g_set_error (error,
- 			     LM_ERROR,                 
- 			     LM_ERROR_CONNECTION_OPEN,   
- 			     "gethostbyname() failed");
-		return FALSE;
-        }
+	memset (&req, 0, sizeof(req));
 
-        haddr = ((struct in_addr *) (he->h_addr_list)[0]);
+	req.ai_family   = AF_UNSPEC;
+	req.ai_socktype = SOCK_STREAM;
+	req.ai_protocol = IPPROTO_TCP;
+
+	g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+	       "Going to connect to %s\n",connection->server);
+	
+	if ((err = getaddrinfo (connection->server, NULL, &req, &ans)) != 0) {
+		g_set_error (error,
+			     LM_ERROR,                 
+			     LM_ERROR_CONNECTION_OPEN,   
+			     "getaddrinfo() failed");
+		return FALSE;
+	}
 
 #ifdef HAVE_GNUTLS
 	if (connection->use_ssl) {
 		gnutls_global_init ();
-		gnutls_certificate_allocate_credentials(&connection->gnutls_xcred);
+		gnutls_certificate_allocate_credentials (&connection->gnutls_xcred);
 	}
 #endif
 
-	fd = socket(AF_INET, SOCK_STREAM, 0);
-	memset(&saddr, 0, sizeof(saddr));
-	memcpy(&saddr.sin_addr, haddr, sizeof(struct in_addr));
-	saddr.sin_family = AF_INET;
-	saddr.sin_port = htons (connection->port);
+	for (tmpaddr = ans ; tmpaddr != NULL ; tmpaddr = tmpaddr->ai_next) {
+		((struct sockaddr_in *) tmpaddr->ai_addr)->sin_port = htons (connection->port);
 
-        if (connect(fd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
-		g_set_error (error,
- 			     LM_ERROR,           
- 			     LM_ERROR_CONNECTION_OPEN,
- 			     "connect() failed");
+		getnameinfo (tmpaddr->ai_addr,
+			     tmpaddr->ai_addrlen,
+			     name,     sizeof (name),
+			     portname, sizeof (portname),
+			     NI_NUMERICHOST | NI_NUMERICSERV);
+
+		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+		      "Trying %s port %s...\n", name, portname);
+
+		fd = socket (tmpaddr->ai_family, 
+			     tmpaddr->ai_socktype, 
+			     tmpaddr->ai_protocol);
+		if (fd < 0) {
+			continue;
+		}
+		
+		err = connect (fd,tmpaddr->ai_addr, tmpaddr->ai_addrlen);
+		if (err == 0) {
+			/* connection successfull */
+			break;
+		}
+		
 		close (fd);
+	}
+	
+	freeaddrinfo (ans);
+
+	/* check for failure */
+	if (fd < 0 || err < 0) {
+		g_set_error (error,
+			     LM_ERROR,
+			     LM_ERROR_CONNECTION_OPEN,
+			     "connection failed");
 		return FALSE;
-        }
+	}
 
 #ifdef HAVE_GNUTLS
 	if (connection->use_ssl) {
@@ -313,18 +368,18 @@ connection_do_open (LmConnection *connection, GError **error)
 	g_io_channel_set_buffered (connection->io_channel, FALSE);
 	g_io_channel_set_flags (connection->io_channel,
 				G_IO_FLAG_NONBLOCK, NULL);
-	g_io_add_watch (connection->io_channel,
-			G_IO_IN,
-			(GIOFunc) connection_in_event,
-			connection);
-	g_io_add_watch (connection->io_channel, 
-			G_IO_ERR,
-			(GIOFunc) connection_error_event,
-			connection);
-	g_io_add_watch (connection->io_channel,
-			G_IO_HUP,
-			(GIOFunc) connection_hup_event,
-			connection);
+	connection->io_watch_in = g_io_add_watch (connection->io_channel,
+						  G_IO_IN,
+						  (GIOFunc) connection_in_event,
+						  connection);
+	connection->io_watch_err = g_io_add_watch (connection->io_channel, 
+						   G_IO_ERR,
+						   (GIOFunc) connection_error_event,
+						   connection);
+	connection->io_watch_hup = g_io_add_watch (connection->io_channel,
+						   G_IO_HUP,
+						   (GIOFunc) connection_hup_event,
+						   connection);
 
 	connection->is_open = TRUE;
 
@@ -341,10 +396,18 @@ static void
 connection_do_close (LmConnection *connection)
 {
 	if (connection->io_channel) {
+		g_source_remove (connection->io_watch_in);
+		g_source_remove (connection->io_watch_err);
+		g_source_remove (connection->io_watch_hup);
+
 		g_io_channel_unref (connection->io_channel);
+		connection->io_channel = NULL;
+	} 
+
+	if (!connection->is_open) {
+		return;
 	}
 
-	connection->io_channel = NULL;
 	connection->is_open = FALSE;
 
 #ifdef HAVE_GNUTLS
@@ -362,39 +425,71 @@ connection_in_event (GIOChannel   *source,
 		     GIOCondition  condition,
 		     LmConnection *connection)
 {
-	gchar             buf[IN_BUFFER_SIZE];
-	gsize             bytes_read;
+	gchar     buf[IN_BUFFER_SIZE];
+	gint      bytes_read;
+	GIOStatus status;
        
 	if (!connection->io_channel) {
 		return FALSE;
 	}
+
 #ifdef HAVE_GNUTLS
 	if (connection->use_ssl) {
 		bytes_read = gnutls_record_recv (connection->gnutls_session,
 						 buf,IN_BUFFER_SIZE - 1);
 		if (bytes_read <= 0) {
-			connection_error_event (connection->io_channel, 
-						G_IO_HUP,
-						connection);
+			status = G_IO_STATUS_ERROR;
+			
+			//connection_error_event (connection->io_channel, 
+			//			G_IO_HUP,
+			//			connection);
+		}
+		else {
+			status = G_IO_STATUS_NORMAL;
 		}
 	} else {
 #endif
-	    g_io_channel_read_chars (connection->io_channel,
-				     buf, IN_BUFFER_SIZE - 1,
-				     &bytes_read,
-				     NULL);
+	    status = g_io_channel_read_chars (connection->io_channel,
+					      buf, IN_BUFFER_SIZE - 1,
+					      &bytes_read,
+					      NULL);
 #ifdef HAVE_GNUTLS
 	}
 #endif
 
+	if (status != G_IO_STATUS_NORMAL) {
+		gint reason;
+		
+		switch (status) {
+		case G_IO_STATUS_EOF:
+			reason = LM_DISCONNECT_REASON_HUP;
+			break;
+		case G_IO_STATUS_AGAIN:
+			return TRUE;
+			break;
+		case G_IO_STATUS_ERROR:
+			reason = LM_DISCONNECT_REASON_ERROR;
+			break;
+		default:
+			reason = LM_DISCONNECT_REASON_UNKNOWN;
+		}
+
+		connection_do_close (connection);
+		connection_signal_disconnect (connection, reason);
+		
+		return FALSE;
+	}
+
 	buf[bytes_read] = '\0';
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV:\n");
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "%s\n", buf);
+	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "'%s'\n", buf);
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
-	
+
+	lm_verbose ("Read: %d chars\n", bytes_read);
+
 	lm_parser_parse (connection->parser, buf);
 	
 	return TRUE;
@@ -405,6 +500,10 @@ connection_error_event (GIOChannel   *source,
 			GIOCondition  condition,
 			LmConnection *connection)
 {
+	if (!connection->io_channel) {
+		return FALSE;
+	}
+
 	lm_verbose ("Error event: %d\n", condition);
 	
 	connection_do_close (connection);
@@ -418,6 +517,10 @@ connection_hup_event (GIOChannel   *source,
 		      GIOCondition  condition,
 		      LmConnection *connection)
 {
+	if (!connection->io_channel) {
+		return FALSE;
+	}
+
 	lm_verbose ("HUP event\n");
 
 	connection_do_close (connection);
@@ -688,7 +791,7 @@ connection_incoming_prepare (GSource *source, gint *timeout)
 	
 	connection = ((LmIncomingSource *)source)->connection;
 	
-	return !g_queue_is_empty (connection->incoming_messages);
+	return !lm_queue_is_empty (connection->incoming_messages);
 }
 
 static gboolean
@@ -707,7 +810,7 @@ connection_incoming_dispatch (GSource *source,
 	
 	connection = ((LmIncomingSource *) source)->connection;
 
-	m = (LmMessage *) g_queue_pop_head (connection->incoming_messages);
+	m = (LmMessage *) lm_queue_pop_head (connection->incoming_messages);
 	
 	if (m) {
 		connection_handle_message (connection, m);
@@ -769,7 +872,7 @@ lm_connection_new (const gchar *server)
 	connection->port              = LM_CONNECTION_DEFAULT_PORT;
 	connection->use_ssl           = FALSE;
 	connection->disconnect_cb     = NULL;
-	connection->incoming_messages = g_queue_new ();
+	connection->incoming_messages = lm_queue_new ();
 	connection->incoming_source   = connection_create_source (connection);
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
@@ -869,7 +972,6 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 	gboolean   result;
 	gboolean   finished = FALSE;
 	gboolean   ret_val = FALSE;
-	gint       last_len = 0;
 
 	g_return_val_if_fail (connection != NULL, FALSE);
 	
@@ -911,23 +1013,25 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 	g_source_unref (connection->incoming_source);
 
 	while (!finished) {
-		gint len;
+		gint n;
 		
 		g_main_context_iteration (NULL, TRUE);
 		
-		len = connection->incoming_messages->length;
-		
-		if (len > last_len) {
+		if (lm_queue_is_empty (connection->incoming_messages)) {
+			continue;
+		}
+
+		for (n = 0; n < connection->incoming_messages->length; n++) {
 			LmMessage *m;
-			
-			last_len = len;
-			m = (LmMessage *) g_queue_peek_tail (connection->incoming_messages);
+
+			m = lm_queue_peek_nth (connection->incoming_messages, n);
 			if (lm_message_get_type (m) == LM_MESSAGE_TYPE_STREAM) {
 				connection->stream_id = 
 					g_strdup (lm_message_node_get_attribute (m->node, "id"));
 				ret_val = TRUE;
 				finished = TRUE;
-				g_queue_pop_tail (connection->incoming_messages);
+				lm_queue_remove_nth (connection->incoming_messages, n);
+				break;
 			}
 		}
 	}
@@ -1348,7 +1452,6 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 {
 	gchar     *id;
 	LmMessage *reply = NULL;
-	gint       last_len = 0;
 
 	if (lm_message_node_get_attribute (message->node, "id")) {
 		id = g_strdup (lm_message_node_get_attribute (message->node, 
@@ -1364,24 +1467,27 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 	lm_connection_send (connection, message, error);
 
 	while (!reply) {
-		gint         len;
 		const gchar *m_id;
-		
+		gint         n;
+
 		g_main_context_iteration (NULL, TRUE);
-		
-		len = connection->incoming_messages->length;
-		
-		if (len > last_len) {
+	
+		if (lm_queue_is_empty (connection->incoming_messages)) {
+			continue;
+		}
+
+		for (n = 0; n < connection->incoming_messages->length; n++) {
 			LmMessage *m;
-			last_len = len;
-			m = (LmMessage *) g_queue_peek_tail (connection->incoming_messages);
+
+			m = lm_queue_peek_nth (connection->incoming_messages, n);
+
 			m_id = lm_message_node_get_attribute (m->node, "id");
 			
 			if (m_id && strcmp (m_id, id) == 0) {
 				reply = m;
+				lm_queue_remove_nth (connection->incoming_messages, n);
+				break;
 			}
-
-			g_queue_pop_tail (connection->incoming_messages);
 		}
 	}
 
